@@ -26,7 +26,6 @@
 #include "soup-socket.h"
 #include "soup-ssl.h"
 #include "soup-uri.h"
-#include "soup-enum-types.h"
 
 typedef struct {
 	SoupSocket  *socket;
@@ -34,13 +33,12 @@ typedef struct {
 	SoupAddress *remote_addr, *tunnel_addr;
 	SoupURI     *proxy_uri;
 	gpointer     ssl_creds;
-	gboolean     ssl_strict;
 
 	GMainContext      *async_context;
 
 	SoupMessage *cur_req;
 	SoupConnectionState state;
-	time_t       unused_timeout;
+	gboolean     ever_used;
 	guint        io_timeout, idle_timeout;
 	GSource     *idle_timeout_src;
 } SoupConnectionPrivate;
@@ -62,12 +60,9 @@ enum {
 	PROP_TUNNEL_ADDRESS,
 	PROP_PROXY_URI,
 	PROP_SSL_CREDS,
-	PROP_SSL_STRICT,
 	PROP_ASYNC_CONTEXT,
 	PROP_TIMEOUT,
 	PROP_IDLE_TIMEOUT,
-	PROP_STATE,
-	PROP_MESSAGE,
 
 	LAST_PROP
 };
@@ -80,11 +75,6 @@ static void get_property (GObject *object, guint prop_id,
 static void stop_idle_timer (SoupConnectionPrivate *priv);
 static void clear_current_request (SoupConnection *conn);
 
-/* Number of seconds after which we close a connection that hasn't yet
- * been used.
- */
-#define SOUP_CONNECTION_UNUSED_TIMEOUT 3
-
 static void
 soup_connection_init (SoupConnection *conn)
 {
@@ -100,8 +90,6 @@ finalize (GObject *object)
 		g_object_unref (priv->remote_addr);
 	if (priv->tunnel_addr)
 		g_object_unref (priv->tunnel_addr);
-	if (priv->proxy_uri)
-		soup_uri_free (priv->proxy_uri);
 
 	if (priv->async_context)
 		g_main_context_unref (priv->async_context);
@@ -176,13 +164,6 @@ soup_connection_class_init (SoupConnectionClass *connection_class)
 				      "SSL credentials",
 				      "Opaque SSL credentials for this connection",
 				      G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
-	g_object_class_install_property (
-		object_class, PROP_SSL_STRICT,
-		g_param_spec_boolean (SOUP_CONNECTION_SSL_STRICT,
-				      "Strictly validate SSL certificates",
-				      "Whether certificate errors should be considered a connection error",
-				      TRUE,
-				      G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
 	g_object_class_install_property (
 		object_class, PROP_ASYNC_CONTEXT,
 		g_param_spec_pointer (SOUP_CONNECTION_ASYNC_CONTEXT,
@@ -203,20 +184,6 @@ soup_connection_class_init (SoupConnectionClass *connection_class)
 				   "Connection lifetime when idle",
 				   0, G_MAXUINT, 0,
 				   G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
-	g_object_class_install_property (
-		object_class, PROP_STATE,
-		g_param_spec_enum (SOUP_CONNECTION_STATE,
-				   "Connection state",
-				   "Current state of connection",
-				   SOUP_TYPE_CONNECTION_STATE, SOUP_CONNECTION_NEW,
-				   G_PARAM_READWRITE));
-	g_object_class_install_property (
-		object_class, PROP_MESSAGE,
-		g_param_spec_object (SOUP_CONNECTION_MESSAGE,
-				     "Message",
-				     "Message being processed",
-				     SOUP_TYPE_MESSAGE,
-				     G_PARAM_READABLE));
 }
 
 
@@ -255,9 +222,6 @@ set_property (GObject *object, guint prop_id,
 	case PROP_SSL_CREDS:
 		priv->ssl_creds = g_value_get_pointer (value);
 		break;
-	case PROP_SSL_STRICT:
-		priv->ssl_strict = g_value_get_boolean (value);
-		break;
 	case PROP_ASYNC_CONTEXT:
 		priv->async_context = g_value_get_pointer (value);
 		if (priv->async_context)
@@ -269,9 +233,6 @@ set_property (GObject *object, guint prop_id,
 	case PROP_IDLE_TIMEOUT:
 		priv->idle_timeout = g_value_get_uint (value);
 		break;
-	case PROP_STATE:
-		soup_connection_set_state (SOUP_CONNECTION (object), g_value_get_uint (value));
-		break;
 	default:
 		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
 		break;
@@ -297,9 +258,6 @@ get_property (GObject *object, guint prop_id,
 	case PROP_SSL_CREDS:
 		g_value_set_pointer (value, priv->ssl_creds);
 		break;
-	case PROP_SSL_STRICT:
-		g_value_set_boolean (value, priv->ssl_strict);
-		break;
 	case PROP_ASYNC_CONTEXT:
 		g_value_set_pointer (value, priv->async_context ? g_main_context_ref (priv->async_context) : NULL);
 		break;
@@ -309,12 +267,6 @@ get_property (GObject *object, guint prop_id,
 	case PROP_IDLE_TIMEOUT:
 		g_value_set_uint (value, priv->idle_timeout);
 		break;
-	case PROP_STATE:
-		g_value_set_enum (value, priv->state);
-		break;
-	case PROP_MESSAGE:
-		g_value_set_object (value, priv->cur_req);
-		break;
 	default:
 		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
 		break;
@@ -351,27 +303,18 @@ stop_idle_timer (SoupConnectionPrivate *priv)
 }
 
 static void
-set_current_request (SoupConnection *conn, SoupMessage *req)
+set_current_request (SoupConnectionPrivate *priv, SoupMessage *req)
 {
-	SoupConnectionPrivate *priv = SOUP_CONNECTION_GET_PRIVATE (conn);
-
 	g_return_if_fail (priv->cur_req == NULL);
 
-	g_object_freeze_notify (G_OBJECT (conn));
-
 	stop_idle_timer (priv);
 
 	soup_message_set_io_status (req, SOUP_MESSAGE_IO_STATUS_RUNNING);
 	priv->cur_req = req;
-	g_object_notify (G_OBJECT (conn), "message");
-
 	if (priv->state == SOUP_CONNECTION_IDLE ||
 	    req->method != SOUP_METHOD_CONNECT)
-		soup_connection_set_state (conn, SOUP_CONNECTION_IN_USE);
-
+		priv->state = SOUP_CONNECTION_IN_USE;
 	g_object_add_weak_pointer (G_OBJECT (req), (gpointer)&priv->cur_req);
-
-	g_object_thaw_notify (G_OBJECT (conn));
 }
 
 static void
@@ -379,17 +322,8 @@ clear_current_request (SoupConnection *conn)
 {
 	SoupConnectionPrivate *priv = SOUP_CONNECTION_GET_PRIVATE (conn);
 
-	g_object_freeze_notify (G_OBJECT (conn));
-
-	priv->unused_timeout = 0;
-
-	if (priv->state == SOUP_CONNECTION_IN_USE) {
-		/* We don't use soup_connection_set_state here since
-		 * it may call clear_current_request()...
-		 */
+	if (priv->state == SOUP_CONNECTION_IN_USE)
 		priv->state = SOUP_CONNECTION_IDLE;
-		g_object_notify (G_OBJECT (conn), "state");
-	}
 	start_idle_timer (conn);
 	if (priv->cur_req) {
 		SoupMessage *cur_req = priv->cur_req;
@@ -397,15 +331,14 @@ clear_current_request (SoupConnection *conn)
 		g_object_remove_weak_pointer (G_OBJECT (priv->cur_req),
 					      (gpointer)&priv->cur_req);
 		priv->cur_req = NULL;
-		g_object_notify (G_OBJECT (conn), "message");
 
 		if (!soup_message_is_keepalive (cur_req))
 			soup_connection_disconnect (conn);
-		else
+		else {
+			priv->ever_used = TRUE;
 			soup_message_io_stop (cur_req);
+		}
 	}
-
-	g_object_thaw_notify (G_OBJECT (conn));
 }
 
 static void
@@ -424,21 +357,7 @@ static void
 socket_connect_result (SoupSocket *sock, guint status, gpointer user_data)
 {
 	SoupConnectionAsyncConnectData *data = user_data;
-	SoupConnectionPrivate *priv;
-
-	if (!data->conn) {
-		if (data->callback) {
-			data->callback (NULL, SOUP_STATUS_CANCELLED,
-					data->callback_data);
-		}
-		g_slice_free (SoupConnectionAsyncConnectData, data);
-		return;
-	}
-
-	g_object_remove_weak_pointer (G_OBJECT (data->conn),
-				      (gpointer *)&data->conn);
-
-	priv = SOUP_CONNECTION_GET_PRIVATE (data->conn);
+	SoupConnectionPrivate *priv = SOUP_CONNECTION_GET_PRIVATE (data->conn);
 
 	if (!SOUP_STATUS_IS_SUCCESSFUL (status))
 		goto done;
@@ -453,8 +372,7 @@ socket_connect_result (SoupSocket *sock, guint status, gpointer user_data)
 	g_signal_connect (priv->socket, "disconnected",
 			  G_CALLBACK (socket_disconnected), data->conn);
 
-	soup_connection_set_state (data->conn, SOUP_CONNECTION_IDLE);
-	priv->unused_timeout = time (NULL) + SOUP_CONNECTION_UNUSED_TIMEOUT;
+	priv->state = SOUP_CONNECTION_IDLE;
 	start_idle_timer (data->conn);
 
  done:
@@ -486,18 +404,16 @@ soup_connection_connect_async (SoupConnection *conn,
 	priv = SOUP_CONNECTION_GET_PRIVATE (conn);
 	g_return_if_fail (priv->socket == NULL);
 
-	soup_connection_set_state (conn, SOUP_CONNECTION_CONNECTING);
+	priv->state = SOUP_CONNECTION_CONNECTING;
 
 	data = g_slice_new (SoupConnectionAsyncConnectData);
 	data->conn = conn;
 	data->callback = callback;
 	data->callback_data = user_data;
-	g_object_add_weak_pointer (G_OBJECT (conn), (gpointer *)&data->conn);
 
 	priv->socket =
 		soup_socket_new (SOUP_SOCKET_REMOTE_ADDRESS, priv->remote_addr,
 				 SOUP_SOCKET_SSL_CREDENTIALS, priv->ssl_creds,
-				 SOUP_SOCKET_SSL_STRICT, priv->ssl_strict,
 				 SOUP_SOCKET_ASYNC_CONTEXT, priv->async_context,
 				 SOUP_SOCKET_TIMEOUT, priv->io_timeout,
 				 NULL);
@@ -523,12 +439,11 @@ soup_connection_connect_sync (SoupConnection *conn)
 	priv = SOUP_CONNECTION_GET_PRIVATE (conn);
 	g_return_val_if_fail (priv->socket == NULL, SOUP_STATUS_MALFORMED);
 
-	soup_connection_set_state (conn, SOUP_CONNECTION_CONNECTING);
+	priv->state = SOUP_CONNECTION_CONNECTING;
 
 	priv->socket =
 		soup_socket_new (SOUP_SOCKET_REMOTE_ADDRESS, priv->remote_addr,
 				 SOUP_SOCKET_SSL_CREDENTIALS, priv->ssl_creds,
-				 SOUP_SOCKET_SSL_STRICT, priv->ssl_strict,
 				 SOUP_SOCKET_FLAG_NONBLOCKING, FALSE,
 				 SOUP_SOCKET_TIMEOUT, priv->io_timeout,
 				 NULL);
@@ -549,8 +464,7 @@ soup_connection_connect_sync (SoupConnection *conn)
 	}
 
 	if (SOUP_STATUS_IS_SUCCESSFUL (status)) {
-		soup_connection_set_state (conn, SOUP_CONNECTION_IDLE);
-		priv->unused_timeout = time (NULL) + SOUP_CONNECTION_UNUSED_TIMEOUT;
+		priv->state = SOUP_CONNECTION_IDLE;
 		start_idle_timer (conn);
 	} else {
 	fail:
@@ -606,7 +520,6 @@ soup_connection_disconnect (SoupConnection *conn)
 	g_return_if_fail (SOUP_IS_CONNECTION (conn));
 	priv = SOUP_CONNECTION_GET_PRIVATE (conn);
 
-	soup_connection_set_state (conn, SOUP_CONNECTION_DISCONNECTED);
 	if (!priv->socket)
 		return;
 
@@ -616,6 +529,52 @@ soup_connection_disconnect (SoupConnection *conn)
 	g_object_unref (priv->socket);
 	priv->socket = NULL;
 
+	/* Don't emit "disconnected" if we aren't yet connected */
+	if (priv->state < SOUP_CONNECTION_IDLE)
+		return;
+
+	priv->state = SOUP_CONNECTION_DISCONNECTED;
+
+	if (priv->cur_req &&
+	    priv->cur_req->status_code == SOUP_STATUS_IO_ERROR &&
+	    priv->ever_used) {
+		/* There was a message queued on this connection, but
+		 * the socket was closed while it was being sent.
+		 * Since ever_used is TRUE, then that means at least
+		 * one message was successfully sent on this
+		 * connection before, and so the most likely cause of
+		 * the IO_ERROR is that the connection was idle for
+		 * too long and the server timed out and closed it
+		 * (and we didn't notice until after we started
+		 * sending the message). So we want the message to get
+		 * tried again on a new connection. The only code path
+		 * that could have gotten us to this point is through
+		 * the call to io_cleanup() in
+		 * soup_message_io_finished(), and so all we need to
+		 * do to get the message requeued in this case is to
+		 * change its status.
+		 */
+		soup_message_cleanup_response (priv->cur_req);
+		soup_message_set_io_status (priv->cur_req,
+					    SOUP_MESSAGE_IO_STATUS_QUEUED);
+	}
+
+	/* If cur_req is non-NULL but priv->ever_used is FALSE, then that
+	 * means this was the first message to be sent on this
+	 * connection, and it failed, so the error probably means that
+	 * there's some network or server problem, so we let the
+	 * IO_ERROR be returned to the caller.
+	 *
+	 * (Of course, it's also possible that the error in the
+	 * ever_used == TRUE case was because of a network/server problem
+	 * too. It's even possible that the message crashed the
+	 * server. In this case, requeuing it was the wrong thing to
+	 * do, but presumably, the next attempt will also get an
+	 * error, and eventually the message will be requeued onto a
+	 * fresh connection and get an error, at which point the error
+	 * will finally be returned to the caller.)
+	 */
+
 	/* NB: this might cause conn to be destroyed. */
 	g_signal_emit (conn, signals[DISCONNECTED], 0);
 }
@@ -653,12 +612,9 @@ soup_connection_get_state (SoupConnection *conn)
 		pfd.events = G_IO_IN;
 		pfd.revents = 0;
 		if (g_poll (&pfd, 1, 0) == 1)
-			soup_connection_set_state (conn, SOUP_CONNECTION_REMOTE_DISCONNECTED);
+			priv->state = SOUP_CONNECTION_REMOTE_DISCONNECTED;
 	}
 #endif
-	if (priv->state == SOUP_CONNECTION_IDLE &&
-	    priv->unused_timeout && priv->unused_timeout < time (NULL))
-		soup_connection_set_state (conn, SOUP_CONNECTION_REMOTE_DISCONNECTED);
 
 	return priv->state;
 }
@@ -666,29 +622,13 @@ soup_connection_get_state (SoupConnection *conn)
 void
 soup_connection_set_state (SoupConnection *conn, SoupConnectionState state)
 {
-	SoupConnectionPrivate *priv;
-	SoupConnectionState old_state;
-
 	g_return_if_fail (SOUP_IS_CONNECTION (conn));
-	g_return_if_fail (state >= SOUP_CONNECTION_NEW &&
-			  state <= SOUP_CONNECTION_DISCONNECTED);
+	g_return_if_fail (state > SOUP_CONNECTION_NEW &&
+			  state < SOUP_CONNECTION_DISCONNECTED);
 
-	priv = SOUP_CONNECTION_GET_PRIVATE (conn);
-	old_state = priv->state;
-	priv->state = state;
-	if (state == SOUP_CONNECTION_IDLE &&
-	    old_state == SOUP_CONNECTION_IN_USE)
+	SOUP_CONNECTION_GET_PRIVATE (conn)->state = state;
+	if (state == SOUP_CONNECTION_IDLE)
 		clear_current_request (conn);
-
-	g_object_notify (G_OBJECT (conn), "state");
-}
-
-gboolean
-soup_connection_get_ever_used (SoupConnection *conn)
-{
-	g_return_val_if_fail (SOUP_IS_CONNECTION (conn), FALSE);
-
-	return SOUP_CONNECTION_GET_PRIVATE (conn)->unused_timeout == 0;
 }
 
 /**
@@ -710,7 +650,7 @@ soup_connection_send_request (SoupConnection *conn, SoupMessage *req)
 	g_return_if_fail (priv->state != SOUP_CONNECTION_NEW && priv->state != SOUP_CONNECTION_DISCONNECTED);
 
 	if (req != priv->cur_req)
-		set_current_request (conn, req);
+		set_current_request (priv, req);
 	soup_message_send_request (req, priv->socket, conn,
 				   priv->proxy_uri != NULL);
 }
